"""Graphette: Step-by-step graph algorithm visualization."""

from graphette.models import Config, h, snap
from graphette.visualize import animate, visualize

__version__ = "0.1.1"
__all__ = ["Config", "visualize", "animate"]



from collections import defaultdict, deque

from graphette.models import h, snap


def adjacency(edges, digraph=False):
    adj = defaultdict(list)
    for u, v in edges:
        adj[u].append(v)
        if not digraph:
            adj[v].append(u)
    return adj


def adj_remove_edge(adj, u, v):
    adj[u].remove(v)
    adj[v].remove(u)


def adj_add_edge(adj, u, v):
    adj[u].append(v)
    adj[v].append(u)


def adj_dfs_count(adj, v, visited):
    visited.add(v)
    for u in adj[v]:
        if u not in visited:
            adj_dfs_count(adj, u, visited)


def bfs(edges, start, digraph=False):
    adj = adjacency(edges, digraph=digraph)
    q, v = deque([start]), set([start])

    yield snap(h("g", "visited", n=v), e=edges, ev="init",
               s={'queue': list(q), 'visited': v})

    while q:
        cur = q.popleft()
        yield snap(h("g", "visited", n=v), h("orange", "current", n=[cur]),
                   ev="dequeue", e=edges, s={'queue': list(q), 'visited': list(v)})

        for neigh in adj[cur]:
            if neigh in v:
                continue

            v.add(neigh)
            q.append(neigh)
            yield snap(h("g", "visited", n=v), h("r", "discovered", n=[neigh]),
                       ev="discover", e=edges, s={'queue': list(q), 'visited': v})


def _is_valid_next_edge(adj, u, v):
    if len(adj[u]) == 1:
        return True

    visited_before = set()
    adj_dfs_count(adj, u, visited_before)

    adj_remove_edge(adj, u, v)
    visited_after = set()
    adj_dfs_count(adj, u, visited_after)

    adj_add_edge(adj, u, v)
    return len(visited_before) == len(visited_after)


def fleury(edges):
    adj = adjacency(edges)

    odd_vertices = [v for v in adj if len(adj[v]) % 2 == 1]
    start = odd_vertices[0] if odd_vertices else next(iter(adj))

    yield snap(ev="init", e=edges)

    path, u = [start], start
    while any(adj.values()):
        pe = [(path[i], path[i + 1]) for i in range(len(path) - 1)]
        yield snap(h("b", "path", n=path, e=pe), ev="traverse", e=edges)

        for v in list(adj[u]):
            yield snap(h("b", "path", n=path, e=pe), h("m", "current", n=[v], e=[(u, v)]),
                       e=edges, ev="check_edge")

            if _is_valid_next_edge(adj, u, v):
                adj_remove_edge(adj, u, v)
                path.append(v)
                u = v
                break



from copy import deepcopy
from dataclasses import dataclass, field


@dataclass
class Highlight:
    label: str
    color: str
    nodes: list = field(default_factory=list)
    edges: list = field(default_factory=list)


@dataclass
class Snapshot:
    nodes: list = field(default_factory=list)
    edges: list = field(default_factory=list)
    event: str | None = None
    state: dict = field(default_factory=dict)
    highlights: list = field(default_factory=list)

    def node_colors(self, all_nodes, default="lightgray"):
        node_map = {}
        for h in self.highlights:
            for n in h.nodes:
                node_map[n] = h.color
        return [node_map.get(n, default) for n in all_nodes]

    def edge_colors(self, all_edges, default="black", digraph=False):
        edge_map = {}
        for h in self.highlights:
            for e in h.edges:
                edge_map[e] = h.color

        def _get_color(e):
            if e in edge_map: return edge_map[e]
            if not digraph and (e[1], e[0]) in edge_map:
                return edge_map[(e[1], e[0])]
            return default

        return [_get_color(e) for e in all_edges]

    def legend_entries(self):
        return {h.color: h.label for h in self.highlights if h.label}


def snap(*highlights, ev=None, n=None, e=None, s=None):
    return Snapshot(
        nodes=deepcopy(n or []),
        edges=deepcopy(e or []),
        event=ev,
        state=deepcopy(s or {}),
        highlights=deepcopy(highlights or []),
    )


def h(c, lb=None, *, n=None, e=None):
    return Highlight(label=lb, color=c, nodes=list(n or []), edges=list(e or []))


@dataclass
class Config:
    @staticmethod
    def draw_factory():
        return dict(node_size=1200, edgecolors="k", width=2, linewidths=2)

    @staticmethod
    def handle_factory():
        return dict(marker="o", markersize=10)

    draw_kwargs: dict = field(default_factory=draw_factory)
    handle_kwargs: dict = field(default_factory=handle_factory)
    fig_kwargs: dict = field(default_factory=dict)
    layout_kwargs: dict = field(default_factory=dict)
    seed: int | None = None
    digraph: bool = False




import matplotlib.animation as animation
import matplotlib.pyplot as plt
import networkx as nx
from toolz import curry

from graphette.models import Config


def to_nx(state, digraph=False):
    G = nx.DiGraph() if digraph else nx.Graph()
    G.add_nodes_from(state.nodes)
    G.add_edges_from(state.edges)
    return G


@curry
def draw(config, pos, states, ax, idx):
    ax.clear()
    s = states[idx]
    G = to_nx(s, digraph=config.digraph)

    node_colors = s.node_colors(G.nodes())
    edge_colors = s.edge_colors(G.edges(), digraph=config.digraph)

    nx.draw_networkx(
        G, pos=pos, ax=ax,
        node_color=node_colors,
        edge_color=edge_colors,
        **config.draw_kwargs
    )

    _draw_legend(config, ax, s)
    _draw_title(ax, s, idx)
    ax.axis("off")
    ax.figure.tight_layout()
    ax.figure.canvas.draw_idle()


def _draw_legend(config, ax, state):
    handles = [
        plt.Line2D([0], [0], color="w", markerfacecolor=color,
                   label=label, **config.handle_kwargs)
        for color, label in state.legend_entries().items()
    ]
    ax.legend(handles=handles, loc="upper right")


def _draw_title(ax, state, idx):
    title = f"Step {idx + 1}"
    if state.event:
        title += f" - {state.event}"
    if state.state:
        state_str = ", ".join(f"{k}: {v}" for k, v in state.state.items())
        title += f"\n{state_str}"
    ax.set_title(title)


def visualize(states, *, config: Config = None):
    config = config or Config()
    states = list(states)

    G_init = to_nx(states[0], digraph=config.digraph)
    pos = nx.spring_layout(G_init, seed=config.seed, **config.layout_kwargs)

    fig, ax = plt.subplots(**config.fig_kwargs)
    draw_fn = draw(config, pos, states)

    current_idx = 0
    draw_fn(ax, current_idx)

    def on_key(event):
        nonlocal current_idx
        delta = {'right': 1, 'left': -1}.get(event.key, 0)
        current_idx = (current_idx + delta) % len(states)
        draw_fn(ax, current_idx)

    fig.canvas.mpl_connect("key_press_event", on_key)
    return fig, draw_fn


def animate(states, filename, *, writer=None, config=None, kwargs=None):
    writer = writer or animation.PillowWriter(fps=1)
    fig, renderer = visualize(states, config=config)
    ani = animation.FuncAnimation(
        fig, renderer, frames=len(renderer.states), **(kwargs or {})
    )
    ani.save(filename, writer=writer)


if __name__ == "__main__":
    from graphette.algo import bfs

    edges = [('A', 'B'), ('A', 'C'),
             ('B', 'D'), ('B', 'E'),
             ('F', 'C'), ('F', 'E')]
    states = bfs(edges, 'A')

    fig, _ = visualize(states, config=Config(seed=42))
    plt.show()
